@@ -123,5 +123,5 @@ def test_atomfind(weights_, nb_classes, coord_expected):
     test_img_ = np.load(test_img)
     coordinates_expected = np.load(coord_expected)
     model_ = load_weights(models.dilUnet(nb_classes), weights_)
-    _, (nn_output, coordinates_) = atomnet.predictor(test_img_, model_).run()
+    _, (nn_output, coordinates_) = atomnet.predictor(model_).run(test_img_)
     assert_allclose(coordinates_[0], coordinates_expected)
\ No newline at end of file
